@@ -11,9 +11,12 @@ import io
 import shutil
 from collections import defaultdict
 import re
+import os
 import multiprocessing as mp
 from tqdm.auto import tqdm
 import warnings
+import csv
+import json
 
 from .twitter import Twitter
 from .article import process_article_url
@@ -25,6 +28,10 @@ from .dgl import build_dgl_dataset
 logger = logging.getLogger(__name__)
 
 
+# Disable tokenizer parallelism
+os.environ['TOKENIZERS_PARALLELISM'] = 'false'
+
+
 # Allows progress bars with `pd.DataFrame.progress_apply`
 tqdm.pandas()
 
@@ -134,6 +141,7 @@ class MuminDataset:
                                                   'distilled-patch16-224'),
                  dataset_dir: Union[str, Path] = './mumin',
                  verbose: bool = False):
+        self.compiled = False
         self._twitter = Twitter(twitter_bearer_token=twitter_bearer_token)
         self.size = size
         self.include_replies = include_replies
@@ -157,8 +165,7 @@ class MuminDataset:
             logger.setLevel(logging.WARNING)
 
         # Raise warning
-        warnings.warn('This dataset is currently under review at NeuRIPS 2021 '
-                      'Datasets and Benchmarks Track (Round 2). This dataset '
+        warnings.warn('This dataset is currently under review. This dataset '
                       'must not be used until this warning is removed, as the '
                       'dataset is subject to change, for example, during the '
                       'review period.')
@@ -170,14 +177,14 @@ class MuminDataset:
             str: The representation of the dataset.
         '''
         if len(self.nodes) == 0 or len(self.rels) == 0:
-            return f'MuminDataset(size={self.size}, compiled=False)'
+            return f'MuminDataset(size={self.size}, compiled={self.compiled})'
         else:
             num_nodes = sum([len(df) for df in self.nodes.values()])
             num_rels = sum([len(df) for df in self.rels.values()])
             return (f'MuminDataset(num_nodes={num_nodes:,}, '
                     f'num_relations={num_rels:,}, '
                     f'size=\'{self.size}\', '
-                    f'compiled=False)')
+                    f'compiled={self.compiled})')
 
     def compile(self, overwrite: bool = False):
         '''Compiles the dataset.
@@ -209,6 +216,9 @@ class MuminDataset:
             self._remove_islands()
             self._dump_to_csv()
 
+        # Mark dataset as compiled
+        self.compiled = True
+
         return self
 
     def _download(self, overwrite: bool = False):
@@ -256,6 +266,9 @@ class MuminDataset:
 
         logger.info('Loading dataset')
 
+        # Set up Pandas `read_csv` parameters
+        csv_params = dict(engine='python', on_bad_lines='skip')
+
         # Reset `nodes` and `relations` to ensure a fresh start
         self.nodes = dict()
         self.rels = dict()
@@ -268,7 +281,7 @@ class MuminDataset:
 
             # Node case: no underscores in file name
             if len(fname.split('_')) == 1:
-                self.nodes[fname] = pd.read_csv(path)
+                self.nodes[fname] = pd.read_csv(path, **csv_params)
 
             # Relation case: exactly two underscores in file name
             elif len(fname.split('_')) > 2:
@@ -276,7 +289,7 @@ class MuminDataset:
                 src = splits[0]
                 tgt = splits[-1]
                 rel = '_'.join(splits[1:-1])
-                self.rels[(src, rel, tgt)] = pd.read_csv(path)
+                self.rels[(src, rel, tgt)] = pd.read_csv(path, **csv_params)
 
             # Otherwise raise error
             else:
@@ -443,6 +456,9 @@ class MuminDataset:
             # Get the tweet IDs
             tweet_ids = self.nodes[node_type].tweet_id.tolist()
 
+            # Store any features the nodes might have had before hydration
+            prehydration_df = self.nodes[node_type].copy()
+
             # Rehydrate the tweets
             tweet_dfs = self._twitter.rehydrate_tweets(tweet_ids=tweet_ids)
 
@@ -460,6 +476,12 @@ class MuminDataset:
                 user_df = tweet_dfs['users']
             self.nodes['user'] = user_df
 
+            # Add prehydration tweet features back to the tweets
+            self.nodes[node_type] = pd.merge(left=self.nodes[node_type],
+                                             right=prehydration_df,
+                                             how='left',
+                                             on='tweet_id')
+
             # Extract and store images
             if self.include_images and len(tweet_dfs['media']):
                 image_df = (tweet_dfs['media']
@@ -1382,7 +1404,7 @@ class MuminDataset:
         # Define embedding function
         def embed(text: str):
             '''Extract a text embedding'''
-            return np.asarray(pipeline(text))[0, 0, :]
+            return np.asarray(pipeline(text, truncation=True))[0, 0, :]
 
         # Embed tweet text using the pretrained transformer
         text_embs = self.nodes['tweet'].text.progress_apply(embed)
@@ -1411,7 +1433,7 @@ class MuminDataset:
         # Define embedding function
         def embed(text: str):
             '''Extract a text embedding'''
-            return np.asarray(pipeline(text))[0, 0, :]
+            return np.asarray(pipeline(text, truncation=True))[0, 0, :]
 
         # Embed tweet text using the pretrained transformer
         text_embs = self.nodes['reply'].text.progress_apply(embed)
@@ -1546,11 +1568,32 @@ class MuminDataset:
         '''Embeds all the claims in the dataset'''
         logger.info('Embedding claims')
 
-        # Embed claim reviewer using a one-hot encoding
-        reviewers = self.nodes['claim'].reviewer.tolist()
+        # Ensure that `reviewers` is a list
+        if isinstance(self.nodes['claim'].reviewers.iloc[0], str):
+
+            def string_to_list(string: str) -> list:
+                string = string.replace('\'', '\"')
+                return json.loads(string)
+
+            self.nodes['claim']['reviewers'] = (self.nodes['claim']
+                                                    .reviewers
+                                                    .map(string_to_list))
+
+        # Set up one-hot encoding of claim reviewers
+        reviewers = self.nodes['claim'].reviewers.explode().unique().tolist()
         one_hotted = [np.asarray(lst)
                       for lst in pd.get_dummies(reviewers).to_numpy().tolist()]
-        self.nodes['claim']['reviewer_emb'] = one_hotted
+        one_hot_dict = {reviewer: array
+                        for reviewer, array in zip(reviewers, one_hotted)}
+
+        def embed_reviewers(revs: List[str]) -> np.ndarray:
+            '''One-hot encoding of multiple reviewers'''
+            arrays = [one_hot_dict[rev] for rev in revs]
+            return np.stack(arrays, axis=0).sum(axis=0)
+
+        # Embed claim reviewer using a one-hot encoding
+        reviewer_emb = self.nodes['claim'].reviewers.map(embed_reviewers)
+        self.nodes['claim']['reviewer_emb'] = reviewer_emb
 
         return self
 
@@ -1559,12 +1602,18 @@ class MuminDataset:
         logger.info('Filters node features')
 
         # Set up the node features that should be kept
-        node_feats = dict(claim=['raw_verdict', 'predicted_verdict',
-                                 'reviewer', 'date'],
+        size = 'small' if self.size == 'test' else self.size
+        node_feats = dict(claim=['embedding', 'label', 'reviewers', 'date',
+                                 f'{size}_train_mask',
+                                 f'{size}_val_mask',
+                                 f'{size}_test_mask'],
                           tweet=['tweet_id', 'text', 'created_at', 'lang',
                                  'source', 'public_metrics.retweet_count',
                                  'public_metrics.reply_count',
-                                 'public_metrics.quote_count'],
+                                 'public_metrics.quote_count', 'label',
+                                 f'{size}_train_mask',
+                                 f'{size}_val_mask',
+                                 f'{size}_test_mask'],
                           reply=['tweet_id', 'text', 'created_at', 'lang',
                                  'source', 'public_metrics.retweet_count',
                                  'public_metrics.reply_count',
@@ -1593,6 +1642,9 @@ class MuminDataset:
             'public_metrics.following_count': 'num_followees',
             'public_metrics.tweet_count': 'num_tweets',
             'public_metrics.listed_count': 'num_listed',
+            f'{size}_train_mask': 'train_mask',
+            f'{size}_val_mask': 'val_mask',
+            f'{size}_test_mask': 'test_mask'
         }
 
         # Filter and rename the node features
@@ -1604,7 +1656,7 @@ class MuminDataset:
                                  for old, new in node_feat_renaming.items()
                                  if old in features}
                 self.nodes[node_type] = (self.nodes[node_type][filtered_feats]
-                                         .rename(columns=renaming_dict))
+                                             .rename(columns=renaming_dict))
 
         return self
 
@@ -1764,13 +1816,17 @@ class MuminDataset:
         for node_type in self._node_dump:
             if node_type in self.nodes.keys():
                 path = self.dataset_dir / f'{node_type}.csv'
-                self.nodes[node_type].to_csv(path, index=False)
+                self.nodes[node_type].to_csv(path,
+                                             index=False,
+                                             quoting=csv.QUOTE_NONNUMERIC)
 
         # Dump the relations
         for rel_type in self._rel_dump:
             if rel_type in self.rels.keys():
                 path = self.dataset_dir / f'{"_".join(rel_type)}.csv'
-                self.rels[rel_type].to_csv(path, index=False)
+                self.rels[rel_type].to_csv(path,
+                                           index=False,
+                                           quoting=csv.QUOTE_NONNUMERIC)
 
         return self
 
